# -*- coding: utf-8 -*-
"""
:Module:            khoros.structures.communities
:Synopsis:          This module contains functions relating to structures (i.e. categories, nodes and tenants)
:Usage:             ``from khoros.structures import base``
:Example:           ``details = base.get_details(khoros_object, 'category', 'category-id')``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     10 Mar 2021
"""

from .. import liql, errors
from ..utils import log_utils
from ..utils.core_utils import display_warning

# Initialize the logger for this module
logger = log_utils.initialize_logging(__name__)


def get_details(khoros_object, identifier='', structure_type=None, first_item=None, community=False):
    """This function retrieves all details for a structure type via LiQL.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param identifier: The identifier (Category/Node ID or URL) by which to filter the results in the ``WHERE`` clause.
    :type identifier: str
    :param structure_type: Designates the structure as a ``category``, ``node`` or ``community``

                           .. note:: Optional if the ``identifier`` is a URL or the ``community`` Boolean is ``True``

    :param first_item: Filters the response data to the first item returned (``True`` by default)
    :type first_item: bool
    :param community: Alternate way of defining the structure type as a ``community`` (``False`` by default)
    :type community: bool
    :returns: The details for the structure type as a dictionary
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    if first_item is not False and (community is False and structure_type != 'community'):
        first_item = True
    if community or structure_type == 'community':
        liql_table = Mapping.structure_types_to_tables.get('community')
        first_item = False if first_item is not True else True
    elif structure_type not in Mapping.structure_types and '/' not in identifier:
        if structure_type:
            raise errors.exceptions.InvalidStructureTypeError(val=structure_type)
        else:
            raise errors.exceptions.MissingRequiredDataError(
                "A structure type (e.g. 'category' or 'node') must be supplied if a " +
                "full URL is not passed as an identifier.")
    elif '/' in identifier:
        if not structure_type:
            structure_type = get_structure_type_from_url(identifier, ignore_exceptions=True)
            if not structure_type:
                raise errors.exceptions.InvalidStructureTypeError("No structure type was provided and unable to " +
                                                                  "define a structure type using the provided URL.")
        liql_table = Mapping.structure_types_to_tables.get(structure_type)
    else:
        liql_table = Mapping.structure_types_to_tables.get(structure_type)
    is_href = True if ('/' in identifier and structure_type != 'node') else False
    where_filter = {True: 'view_href', False: 'id'}
    if '/' in identifier and structure_type == 'node':
        identifier = get_structure_id(identifier)
    try:
        # TODO: Update the query below to be less greedy or at least to provide the option to define fields
        query = f'SELECT * FROM {liql_table}'       # nosec
        if not community and structure_type != 'community':
            query = f'{query} WHERE {where_filter.get(is_href)} = "{identifier}"'
        response = liql.perform_query(khoros_object, liql_query=query, verify_success=True)
    except NameError:
        raise errors.exceptions.MissingRequiredDataError("The LiQL table was not defined")
    if first_item:
        response = response['data']['items'][0]
    return response


def structure_exists(khoros_object, structure_type, structure_id=None, structure_url=None):
    """This function checks to see if a structure (i.e. node, board, category or group hub) exists.

    .. versionadded:: 2.7.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param structure_type: The type of structure (e.g. ``board``, ``category``, ``node`` or ``grouphub``)
                           .. note:: The ``group hub`` value (as two words) is also acceptable.
    :type structure_type: str
    :param structure_id: The ID of the structure to check
    :type structure_id: str, None
    :param structure_url: The URL of the structure to check
    :type structure_url: str, None
    :returns: Boolean value indicating whether or not the structure already exists
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    if not any((structure_id, structure_url)):
        raise errors.exceptions.MissingRequiredDataError("Must provide at least one lookup value.")
    if structure_type not in Mapping.structure_types_to_tables.values():
        if structure_type not in Mapping.structure_types_to_tables.keys():
            raise errors.exceptions.InvalidStructureTypeError(val=structure_type)
        else:
            structure_type = Mapping.structure_types_to_tables.get(structure_type)
    if not structure_id:
        structure_id = get_structure_id(structure_url)
    count = liql.get_total_count(khoros_object, structure_type, f'id = "{structure_id}"')
    return True if count > 0 else False


def get_structure_id(url):
    """This function retrieves the Node ID from a full URL.

    .. versionchanged:: 2.6.0
       The function was renamed from ``_get_node_id`` to ``get_structure_id`` and converted from private to public.

    .. versionadded:: 2.1.0

    :param url: The full URL of the node
    :type url: str
    :returns: The ID in string format
    """
    node_id = ''
    for node_url_code in Mapping.node_url_identifiers:
        if node_url_code in url:
            node_id = url.split(node_url_code)[1]
            break
    if not node_id:
        raise errors.exceptions.InvalidURLError(f"Unable to identify the Node ID from the following URL: {url}")
    return node_id


def _check_url_for_identifier(_url, _id_type, _ignore_exceptions=False):
    """This function checks a URL to see if it has a particular identifier associated with a category or node.

    .. versionadded:: 2.1.0

    :param _url: The URL to be evaluated
    :type _url: str
    :param _id_type: The type of identifier (e.g. ``category``, ``node``)
    :type _id_type: str
    :param _ignore_exceptions: Determines if exceptions should not be raised
    :type _ignore_exceptions: bool
    :returns: Boolean value indicating if a match was found
    :raises: :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`
    """
    if _id_type not in Mapping.structure_types and not _ignore_exceptions:
        raise errors.exceptions.InvalidStructureTypeError(val=_id_type)
    _match_found = False
    _id_lists = {
        'category': Mapping.category_url_identifiers,
        'node': Mapping.node_url_identifiers
    }
    for _identifier in _id_lists.get(_id_type):
        if _identifier in _url:
            _match_found = True
            break
    return _match_found


def limit_to_first_child(structure_data):
    """This function limits structure data to be only for the first child.

    .. versionadded:: 2.1.0

    :param structure_data: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type structure_data: dict
    :returns: The data dictionary that has been restricted to the first child
    :raises: :py:exc:`TypeError`, :py:exc:`KeyError`
    """
    return structure_data['data']['items'][0]


def get_structure_field(khoros_object, field, identifier='', details=None,
                        structure_type=None, first_item=True, community=False):
    """This function returns a specific API field value for a community, category or node collection.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param field: The field from the :py:class:`khoros.structures.base.Mapping` class whose value should be returned
    :type field: str
    :param identifier: The identifier (Category/Node ID or URL) by which to filter the results in the ``WHERE`` clause.
    :type identifier: str
    :type details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type details: dict, None
    :param structure_type: Designates the structure as a ``category``, ``node`` or ``community``

                           .. note:: Optional if the ``identifier`` is a URL or the ``community`` Boolean is ``True``

    :type structure_type: str, None
    :param first_item: Filters the response data to the first item returned (``True`` by default)
    :type first_item: bool
    :param community: Alternate way of defining the structure type as a ``community`` (``False`` by default)
    :type community: bool
    :returns: The API field value in its appropriate format
    :raises: :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    if not details:
        details = get_details(khoros_object, identifier, structure_type, first_item, community)
    structure_type = verify_structure_type(identifier, structure_type, community)
    field_dicts = {
        'category': Mapping.category_fields,
        'community': Mapping.community_fields,
        'node': Mapping.node_fields
    }
    data_fields = field_dicts.get(structure_type)
    if field not in data_fields:
        raise errors.exceptions.InvalidFieldError(val=field)
    api_field = data_fields.get(field)
    if len(api_field) == 1:
        return_field = details[api_field[0]]
    elif len(api_field) == 2:
        return_field = details[api_field[0]][api_field[1]]
    else:
        return_field = details[api_field[0]][api_field[1]][api_field[2]]
    return return_field


def is_category_url(url, ignore_exceptions=False):
    """This function identifies if a provided URL is for a category in the environment.

    .. versionadded:: 2.1.0

    :param url: The URL to be evaluated
    :type url: str
    :param ignore_exceptions: Determines if exceptions should not be raised
    :type ignore_exceptions: bool
    :returns: Boolean value indicating if the URL is associated with a category
    :raises: :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`
    """
    return _check_url_for_identifier(url, 'category', ignore_exceptions)


def is_node_url(url, ignore_exceptions=False):
    """This function identifies if a provided URL is for a node in the environment.

    .. versionadded:: 2.1.0

    :param url: The URL to be evaluated
    :type url: str
    :param ignore_exceptions: Determines if exceptions should not be raised
    :type ignore_exceptions: bool
    :returns: Boolean value indicating if the URL is associated with a node
    :raises: :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`
    """
    return _check_url_for_identifier(url, 'node', ignore_exceptions)


def verify_structure_type(identifier, structure_type, community=False):
    """This function verifies the structure type by examining the identifier(s) and provided structure type value.

    .. versionadded:: 2.1.0

    :param identifier: The identifier (Category/Node ID or URL) by which to filter the results in the parent function
    :type identifier: str
    :param structure_type: Designates the structure as a ``category``, ``node`` or ``community``

                           .. note:: Optional if the ``identifier`` is a URL or the ``community`` Boolean is ``True``

    :type structure_type: str
    :param community: Alternate way of defining the structure type as a ``community`` (``False`` by default)
    :type community: bool
    :returns: The appropriately labeled and verified structure type
    :raises: :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    accepted_structure_types = ['category', 'community', 'node']
    if community:
        structure_type = 'community'
    if structure_type:
        if structure_type in accepted_structure_types:
            return structure_type
        elif structure_type not in accepted_structure_types and '/' not in identifier:
            raise errors.exceptions.InvalidStructureTypeError(val=structure_type)
        else:
            display_warning(f"The structure type '{structure_type}' is invalid. Will attempt to identify via URL.")
    structure_type = get_structure_type_from_url(identifier, ignore_exceptions=True)
    if not structure_type:
        raise errors.exceptions.MissingRequiredDataError(
            "A structure type (e.g. 'category' or 'node') must be supplied if a " +
            "full URL is not passed as an identifier.")
    return structure_type


def get_structure_type_from_url(url, ignore_exceptions=False):
    """This function determines if a URL is for a category or node (or neither).

    .. versionadded:: 2.1.0

    :param url: The URL to be evaluated
    :type url: str
    :param ignore_exceptions: Determines if exceptions should not be raised
    :type ignore_exceptions: bool
    :returns: A string with ``category`` or ``node``, or a blank string if neither

    """
    structure_type = ''
    if is_category_url(url, ignore_exceptions):
        structure_type = 'category'
    elif is_node_url(url, ignore_exceptions):
        structure_type = 'node'
    return structure_type


class Mapping:
    """This class contains lists and dictionaries used to map structure data."""
    category_url_identifiers = ['ct-p/']
    node_url_identifiers = ['bg-p/', 'con-p/', 'bd-p/', 'gp-p/', 'idb-p/', 'qa-p/', 'tkb-p/', 'gh-p/', 'ct-p/']
    structure_types = ['category', 'node']
    structure_types_to_tables = {
        'board': 'boards',
        'category': 'categories',
        'community': 'communities',
        'group hub': 'grouphubs',
        'grouphub': 'grouphubs',
        'node': 'nodes'
    }
    category_fields = {
        'type': ('type',),
        'id': ('id',),
        'href': ('href',),
        'view_href': ('view_href',),
        'full_title': ('title',),
        'short_title': ('short_title',),
        'description': ('description',),
        'parent_details': ('parent_category',),
        'parent_type': ('parent_category', 'type'),
        'parent_id': ('parent_category', 'id'),
        'parent_href': ('parent_category', 'href'),
        'parent_view_href': ('parent_category', 'view_href'),
        'root_details': ('root_category',),
        'root_type': ('root_category', 'type'),
        'root_id': ('root_category', 'id'),
        'root_href': ('root_category', 'href'),
        'root_view_href': ('root_category', 'view_href'),
        'ancestors_query': ('ancestor_categories', 'query'),
        'descendants_query': ('descendant_categories', 'query'),
        'children_query': ('child_categories', 'query'),
        'boards_query': ('boards', 'query'),
        'language': ('language',),
        'hidden': ('hidden',),
        'messages_query': ('messages', 'query'),
        'topics_query': ('topics', 'query'),
        'views': ('views',),
        'date_pattern': ('date_pattern',),
        'friendly_date_enabled': ('friendly_date_enabled',),
        'friendly_date_max_age': ('friendly_date_max_age',),
        'skin': ('skin',),
        'depth': ('depth',),
        'position': ('position',),
        'user_context_details': ('user_context',),
        'user_context_type': ('user_context', 'type'),
        'user_context_sort_order': ('user_context', 'sort_order'),
        'user_context_sort_field': ('user_context', 'sort_field'),
        'creation_date': ('creation_date',),
    }
    community_fields = {
        'type': ('type',),
        'id': ('id',),
        'full_title': ('title',),
        'short_title': ('short_title',),
        'description': ('description',),
        'href': ('href',),
        'view_href': ('view_href',),
        'user_context_details': ('user_context',),
        'user_context_type': ('user_context', 'type'),
        'user_context_sort_order': ('user_context', 'sort_order'),
        'user_context_sort_field': ('user_context', 'sort_field'),
        'attachment_max_per_message': ('attachment_max_per_message',),
        'attachment_file_types': ('attachment_file_types',),
        'email_confirmation_required_to_post': ('email_confirmation_required_to_post',),
        'language': ('language',),
        'ooyala_player_branding_id': ('ooyala_player_branding_id',),
        'date_pattern': ('date_pattern',),
        'friendly_date_enabled': ('friendly_date_enabled',),
        'friendly_date_max_age': ('friendly_date_max_age',),
        'skin': ('skin',),
        'web_ui_details': ('web_ui',),
        'web_ui_type': ('web_ui', 'type'),
        'web_ui_sign_out_url': ('web_ui', 'sign_out_url'),
        'web_ui_redirect_param': ('web_ui', 'redirect_param'),
        'web_ui_redirect_reason_param': ('web_ui', 'redirect_reason_param'),
        'top_level_categories_enabled': ('top_level_categories_enabled',),
        'tlc_show_community_node_in_breadcrumb': ('tlc_show_community_node_in_breadcrumb',),
        'tlc_show_breadcrumb_at_top_level': ('tlc_show_breadcrumb_at_top_level',),
        'tlc_set_on_community_page': ('tlc_set_on_community_page',),
        'creation_date': ('creation_date',)
    }
    node_fields = {
        'type': ('type',),
        'id': ('id',),
        'href': ('href',),
        'node_type': ('node_type',),
        'discussion_style': ('conversation_style',),
        'full_title': ('title',),
        'short_title': ('short_title',),
        'description': ('description',),
        'parent_details': ('parent',),
        'parent_type': ('parent', 'type'),
        'parent_id': ('parent', 'id'),
        'parent_href': ('parent', 'href'),
        'root_details': ('root_category',),
        'root_type': ('root_category', 'type'),
        'root_id': ('root_category', 'id'),
        'root_href': ('root_category', 'href'),
        'ancestors_query': ('ancestors', 'query'),
        'avatar_details': ('avatar',),
        'avatar_type': ('avatar', 'type'),
        'avatar_tiny_href': ('avatar', 'tiny_href'),
        'avatar_small_href': ('avatar', 'small_href'),
        'avatar_medium_href': ('avatar', 'medium_href'),
        'avatar_large_href': ('avatar', 'large_href'),
        'creation_date': ('creation_date',),
        'creation_date_friendly': ('creation_date_friendly',),
        'children_query': ('children', 'query'),
        'depth': ('depth',),
        'position': ('position',),
        'hidden': ('hidden',),
        'user_context_details': ('user_context',),
        'user_context_type': ('user_context', 'type'),
        'user_context_sort_order': ('user_context', 'sort_order'),
        'user_context_sort_field': ('user_context', 'sort_field'),
        'messages_query': ('messages', 'query'),
        'topics_query': ('topics', 'query'),
        'views': ('views',)
    }



# -*- coding: utf-8 -*-
"""
:Module:         khoros.utils.tests.test_grouphub_creation
:Synopsis:       This module is used by pytest to verify that the group hub creation process works properly
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  11 Mar 2021
"""

import pytest

from . import resources


def verify_data_fields(payload, data_fields):
    """This function checks a dictionary of data fields and values to ensure they match what is in the payload.

    :param payload: The payload for a new group hub
    :type payload: dict
    :param data_fields: The data fields and corresponding values to check
    :type data_fields: dict
    :returns: Boolean value indicating whether or not the verification checks out
    """
    verified = True
    for field, value in data_fields.items():
        if field not in payload['data']:
            verified = False
            break
        else:
            if payload['data'].get(field) != value:
                verified = False
                break
    return verified


def test_no_arguments():
    """This function tests to ensure that a TypeError is raised if no arguments are passed to the function."""
    with pytest.raises(TypeError):
        grouphubs.structure_payload()
    return


def test_only_id():
    """This function tests to ensure passing only the ID references the ``group_title`` argument in the exception."""
    with pytest.raises(TypeError) as exc:
        grouphubs.structure_payload()
        assert 'group_title' in str(exc.value)      # nosec
    return


# Import modules and initialize the core object
grouphubs, exceptions = resources.import_modules('khoros.structures.grouphubs', 'khoros.errors.exceptions')
khoros = resources.initialize_khoros_object()



# -*- coding: utf-8 -*-
"""
:Module:            khoros.objects.roles
:Synopsis:          This module includes functions that handle roles and permissions.
:Usage:             ``from khoros.objects import roles``
:Example:           ``count = roles.get_total_role_count()``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     08 Apr 2021
"""

from . import users
from .. import api, liql, errors
from ..utils import log_utils, core_utils

# Initialize the logger for this module
logger = log_utils.initialize_logging(__name__)

ROLE_TYPES = {
    'board': 'b',
    'category': 'c',
    'community': 't',
    'global': 't',
    'grouphub': 'g',
    'group_hub': 'g',
    'top_level': 't'
}
NODE_SPECIFIC_ROLES = {'b', 'board', 'c', 'category', 'g', 'grouphub', 'group_hub'}


def _get_role_type_prefix(_role_type=None):
    """This function identifies the appropriate prefix for a given role type. (Top-level prefix returned by default)

    .. versionadded:: 3.5.0

    :param _role_type: The role type associated with a role (e.g. ``board``, ``category``, etc.)
    :type _role_type: str, None
    :returns: The prefix associated with the role type (e.g. ``b``, ``c``, etc.)
    :raises: :py:exc:`TypeError`
    """
    _prefix = 't'
    if _role_type in ROLE_TYPES.values():
        _prefix = _role_type
    elif _role_type in ROLE_TYPES:
        _prefix = ROLE_TYPES.get(_role_type)
    return _prefix


def get_role_id(role_name, scope='community', node_id=None):
    """This function constructs and returns the Role ID associated with a given role name and scope.

    .. versionadded:: 4.0.0

    :param role_name: The name of the role (e.g. ``Administrator``, ``Moderator``, ``Owner``, etc.)
    :type role_name: str
    :param scope: The scope of the role (``community`` by default)
    :type scope: str
    :param node_id: The associated Node ID for any role that does not have a global/community scope.
    :type node_id: str, None
    :returns: The properly constructed Role ID where applicable
    :raises: :py:exc:`khoros.errors.exceptions.InvalidRoleError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    if scope not in ROLE_TYPES:
        raise errors.exceptions.InvalidRoleError(f"The scope '{scope}' is not valid")
    prefix = _get_role_type_prefix(scope)
    if prefix == "t":
        role_id = f"{prefix}:{role_name}"
    else:
        if not node_id:
            raise errors.exceptions.MissingRequiredDataError(param='node_id')
        role_id = f"{prefix}:{node_id}:{role_name}"
    return role_id


def get_total_role_count(khoros_object, return_dict=False, total=True, top_level=False, board=False, category=False,
                         group_hub=False):
    """This function retrieves the total role count for one or more role type(s).

    .. versionadded:: 2.4.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param return_dict: Determines if the data should be returned as a dictionary (``False`` by default)
    :type return_dict: bool
    :param total: Indicates that the total overall role count should be returned (``True`` by default)
    :type total: bool
    :param top_level: Indicates that the total top-level role count should be returned (``False`` by default)
    :type top_level: bool
    :param board: Indicates that the total board-level role count should be returned (``False`` by default)
    :type board: bool
    :param category: Indicates that the total category-level role count should be returned (``False`` by default)
    :type category: bool
    :param group_hub: Indicates that the total group hub-level role count should be returned (``False`` by default)
    :type group_hub: bool
    :returns: The role count(s) as an integer, tuple or dictionary, depending on the arguments supplied
    :raises: :py:exc:`khoros.errors.exceptions.InvalidRoleTypeError`
    """
    response = liql.perform_query(khoros_object, liql_query="SELECT id FROM roles", verify_success=True)
    counts = {
        'total': api.get_results_count(response)
    }
    if not total:
        del counts['total']
    if top_level or board or category or group_hub:
        roles_dict = api.get_items_list(response)
        count_types = {'top_level': top_level, 'board': board, 'category': category, 'group_hub': group_hub}
        for count_type, should_count in count_types.items():
            if should_count:
                counts[count_type] = count_role_types(count_type, roles_dict)
    if not return_dict:
        if len(counts) == 1:
            counts = counts.get(list(counts.keys())[0])
        else:
            counts = tuple(counts.values())
    return counts


def count_role_types(role_type, roles_dict):
    """This function returns the total count for a specific role type.

    .. versionadded:: 2.4.0

    :param role_type: The role type for which to return the count (e.g. ``board``, ``category``, etc.)
    :type role_type: str
    :param roles_dict: Dictionary of the roles for a given Khoros Community environment
    :type roles_dict: dict
    :returns: The total count for the role type as an integer
    :raises: :py:exc:`khoros.errors.exceptions.InvalidRoleTypeError`
    """
    if role_type not in ROLE_TYPES.keys() and role_type not in ROLE_TYPES.values():
        raise errors.exceptions.InvalidRoleTypeError(role_type=role_type)
    elif role_type in ROLE_TYPES:
        role_type = ROLE_TYPES.get(role_type)
    count = 0
    for role in roles_dict:
        if role['id'].startswith(f"{role_type}:"):
            count += 1
    return count


def get_roles_for_user(khoros_object, user_id, fields=None):
    """This function returns all roles associated with a given User ID.

    .. versionchanged:: 3.5.0
       Fields to return in the LiQL query can now be explicitly defined.

    .. versionadded:: 2.4.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param user_id: The User ID for which to retrieve the roles data
    :type user_id: str, int
    :param fields: The field(s) to retrieve from the LiQL query as a string or list

                   .. note:: All fields (i.e. ``SELECT *``) are returned unless fields are explicitly defined.

    :type fields: str, list, tuple, set, None
    :returns: A dictionary with data for each role associated with the given User ID
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    fields = "*" if not fields else liql.parse_select_fields(fields)
    response = liql.perform_query(khoros_object, liql_query=f"SELECT {fields} FROM roles WHERE users.id = '{user_id}'",
                                  verify_success=True)
    return api.get_items_list(response)


def get_users_with_role(khoros_object, fields='login', role_id=None, role_name=None, scope=None, node_id=None,
                        limit_per_query=1000, cursor=None, where_clause=None, users_list=None, simple=False):
    """This function retrieves a list of all users that have a specific role.

    .. versionchanged:: 4.0.0
       The function now leverages a ``while`` loop instead of recursion in order to avoid raising a
       :py:exc:`RecursionError` exception with larger queries.

    .. versionadded:: 3.5.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param fields: One or more fields from the ``Users`` object to return (``login`` field by default)
    :type fields: str, tuple, list, set
    :param role_id: The identifier for the role in ``node_type:node_id:role_name`` format
    :type role_id: str, None
    :param role_name: The simple role name (e.g. ``Administrator``)

                      .. caution:: This option should only be used when the role name is unique within the community.

    :type role_name: str, None
    :param scope: The scope of the role (e.g. ``board``, ``category``, ``community``, ``grouphub``)

                  .. note:: If a value is not supplied and only a role name is defined then the role scope is
                            assumed to be at the ``community`` level. (i.e. global)

    :type scope: str, None
    :param node_id: The Node ID associated with the role (where applicable)

                    .. note:: If a value is not supplied and only a role name is defined then the role scope is
                              assumed to be at the ``community`` level. (i.e. global)

    :type node_id: str, None
    :param limit_per_query: Defines a ``LIMIT`` constraint other than the default ``1000`` limit per LiQL query

                            .. note:: Unless modified by Khoros Support or Professional Services, ``1000`` is the
                                      maximum number of entries that can be returned in a single LiQL query.

    :type limit_per_query: int, str
    :param cursor: Specifies a cursor to be referenced in a LiQL query

                   .. note:: This parameter is intended for use by the function when calling itself recursively to
                             retrieve users that exceed the ``limit_per_query`` value and will not be leveraged
                             directly in standalone function calls.

    :type cursor: str, None
    :param where_clause: Specifies an exact WHERE clause for the query to be performed

                         .. note:: While technically possible to leverage this parameter in function calls, its
                                   primary use is by the function when calling itself recursively to retrieve users
                                   that exceed the ``limit_per_query`` value.

    :type where_clause: str, None
    :param users_list: Provides an existing list of users that is leveraged when the function is called recursively
    :type users_list: list, None
    :param simple: Returns a simple list of the strings or tuples of the value(s) for each user (``False`` by default)
    :type simple: bool
    :returns: A list of users as strings, tuples or dictionaries depending if ``simple`` mode is enabled
    :raises: :py:exc:`khoros.errors.exceptions.DataMismatchError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    # Initialize the users list if not provided
    users_list = [] if not users_list else users_list

    # Define the WHERE clause if not already defined
    if not where_clause:
        # Ensure appropriate data has been provided
        if not role_id and not role_name:
            raise errors.exceptions.MissingRequiredDataError('A role ID or role name must be provided')
        elif role_id and role_name:
            raise errors.exceptions.DataMismatchError('Only a role ID or role name should be provided and not both')
        elif not node_id and scope in NODE_SPECIFIC_ROLES:
            raise errors.exceptions.MissingRequiredDataError('A node ID must be supplied with node-specific roles')

        # Define the constraint
        prefix = _get_role_type_prefix(scope)
        if role_name:
            if scope in NODE_SPECIFIC_ROLES:
                constraint = f"roles.id = '{prefix}:{node_id}:{role_name}'"
            elif scope is not None:
                constraint = f"roles.id = '{prefix}:{role_name}'"
            else:
                constraint = f"roles.name = '{role_name}'"
        elif ':' not in role_id and scope not in NODE_SPECIFIC_ROLES:
            constraint = f"roles.name = '{role_name}'"
        else:
            constraint = f"roles.id = '{role_id}'"

        # Define the full WHERE clause if not already define4d
        if not any((isinstance(limit_per_query, int), isinstance(limit_per_query, str))):
            raise errors.exceptions.InvalidFieldError('Limit per query value must be a number (integer or string)')
        elif int(limit_per_query) < 0:
            raise errors.exceptions.InvalidFieldError('Limit per query value must be a positive number')
        where_clause = f"WHERE {constraint} LIMIT {limit_per_query}"

    # Properly parse the SELECT statement
    fields = liql.parse_select_fields(fields)

    # Perform the first LiQL query and add to the master users list
    user_data, cursor = _query_for_users(khoros_object, fields, where_clause, cursor)
    users_list.extend(user_data)

    # Continue looping as long as a cursor is present
    while cursor:
        user_data, cursor = _query_for_users(khoros_object, fields, where_clause, cursor)
        users_list.extend(user_data)

    # Convert to simple list when requested
    if simple:
        users_list = core_utils.convert_dict_list_to_simple_list(users_list, fields)

    # Return the populated users list
    return users_list


def _query_for_users(_khoros_object, _fields, _where_clause, _cursor):
    """This function performs a LiQL query to retrieve users with a specific role.

    .. versionadded:: 4.0.0

    :param _khoros_object: The core :py:class:`khoros.Khoros` object
    :type _khoros_object: class[khoros.Khoros]
    :param _fields: One or more fields from the ``Users`` object to return (``login`` field by default)
    :type _fields: str, tuple, list, set
    :param _where_clause: Specifies an exact WHERE clause for the query to be performed
    :type _where_clause: str, None
    :param _cursor: Specifies a cursor to be referenced in a LiQL query
    :type _cursor: str, None
    :returns: The API response with the user data and the next LiQL cursor when applicable
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.LiQLParseError`
    """
    # Construct the entire LiQL query
    _cursor = '' if not _cursor else liql.structure_cursor_clause(_cursor)
    _query = f"SELECT {_fields} FROM users {_where_clause} {_cursor}"

    # Perform the API call and retrieve the data
    _response = liql.perform_query(_khoros_object, liql_query=_query)
    _user_data = liql.get_returned_items(_response)

    # Get the cursor when present
    _cursor = None
    if _response.get('data') and _response['data'].get('next_cursor'):
        _cursor = _response['data'].get('next_cursor')

    # Return the user data and cursor
    return _user_data, _cursor


def _validate_node_type(_node_type=None, _param_name='node_type'):
    """This function verifies that a provided node type (i.e. role scope) is valid for role-related functions.

    .. versionadded:: 4.0.0

    :param _node_type: The node type to be validated
    :type _node_type: str, None
    :param _param_name: The name of the parameter in the parent function in which the node type is stored
                        (Default: ``node_type``)

                        .. note:: This value is used when raising the
                                  :py:exc:`khoros.errors.exceptions.MissingRequiredDataError` exception.

    :type _param_name: str
    :returns: The validated node type
    :raises: :py:exc:`khoros.errors.exceptions.InvalidNodeTypeError`
    """
    if not _node_type:
        raise errors.exceptions.MissingRequiredDataError(param=_param_name)
    if _node_type not in ROLE_TYPES:
        raise errors.exceptions.InvalidNodeTypeError()
    _node_type = 'grouphub' if _node_type == 'group_hub' else _node_type
    return _node_type


def _assign_role_with_v1(_khoros_object, _user, _lookup_type, _role, _node=None, _node_type='board', _return_json=True):
    """This function assigns a role to a user via REST API call using the Community API v1.

    .. versionadded:: 4.0.0

    :param _khoros_object: The core :py:class:`khoros.Khoros` object
    :type _khoros_object: class[khoros.Khoros]
    :param _user: The identifier (i.e. ID, login or email) of the user to be assigned to the role
    :type _user: str
    :param _lookup_type: The lookup type for the user identifier (``id``, ``login`` or ``email``)
    :type _lookup_type: str
    :param _role: The name of the role to which the user will be assigned
    :type _role: str, tuple, list, set
    :param _node: The Node ID of the node to which the role is scoped when applicable
    :type _node: str, None
    :param _node_type: The type of node to which the role is scoped (e.g. ``board`` (default), ``category``, etc.)
    :type _node_type: str
    :param _return_json: Determines if the response should be returned as JSON rather than XML (``True`` by default)
    :type _return_json: bool
    :returns: The response of the API call to assign the user to the role
    :raises: :py:exc:`ValueError`, :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.CurrentlyUnsupportedError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`,
             :py:exc:`khoros.errors.exceptions.UnsupportedNodeTypeError`,
             :py:exc:`khoros.errors.exceptions.InvalidNodeTypeError`,
             :py:exc:`khoros.errors.exceptions.POSTRequestError`
    """
    # Ensure the role has been supplied properly
    if not isinstance(_role, str) and len(_role) > 1:
        raise errors.exceptions.CurrentlyUnsupportedError('Only individual roles can currently be assigned with API v1')
    _role = _role[0] if not isinstance(_role, str) else _role

    # Define the API URI
    _user_path = api.get_v1_user_path(user_and_type=(_user, _lookup_type))
    _uri_role_portion = f"/roles/name/{_role}/users/add?role.user={_user_path}"
    if not _node:
        _uri = f"{_khoros_object.core.get('v1_base')}/{_uri_role_portion}"
    else:
        # TODO: Identify the appropriate URI for grouphub membership additions
        if _node_type == 'grouphub':
            raise errors.exceptions.CurrentlyUnsupportedError('Group hub membership should be added via API v2 calls')
        _collection = api.get_v1_node_collection(_node_type)
        _uri = f"{_khoros_object.core.get('v1_base')}/{_collection}/id/{_node}/{_uri_role_portion}"

    # Perform the POST request and return the response
    return api.post_request_with_retries(_uri, return_json=_return_json, khoros_object=_khoros_object)


def _assign_role_with_v2(_khoros_object, _user, _lookup_type, _roles, _node=None, _node_type='board'):
    """This function assigns one or more roles to a user via REST API call using the Community API v2.

    :param _khoros_object:The core :py:class:`khoros.Khoros` object
    :type _khoros_object: class[khoros.Khoros]
    :param _user: The identifier (i.e. ID, login or email) of the user to be assigned to the role
    :type _user: str
    :param _lookup_type: The lookup type for the user identifier (``id``, ``login`` or ``email``)
    :type _lookup_type: str
    :param _roles: The name of the role(s) to which the user will be assigned
    :type _roles: str, tuple, list, set
    :param _node: The Node ID of the node to which the role is scoped when applicable
    :type _node: str, None
    :param _node_type: The type of node to which the role is scoped (e.g. ``board`` (default), ``category``, etc.)
    :type _node_type: str
    :returns: The response of the API call to assign the user to the role
    :raises: :py:exc:`ValueError`, :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.CurrentlyUnsupportedError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`,
             :py:exc:`khoros.errors.exceptions.UnsupportedNodeTypeError`,
             :py:exc:`khoros.errors.exceptions.InvalidNodeTypeError`,
             :py:exc:`khoros.errors.exceptions.PUTRequestError`
    """
    # Get the User ID if not already provided
    _user_id = _user if _lookup_type == 'id' else None
    if not _user_id:
        if _lookup_type == 'email':
            _user_id = users.get_user_id(_khoros_object, email=_user)
        else:
            _user_id = users.get_user_id(_khoros_object, login=_user)

    # Construct the URI for the API call
    _uri = f"{_khoros_object.core.get('v2_base')}/users/{_user_id}"

    # Define the roles_to_add list within the payload
    _roles_to_add = []
    if isinstance(_roles, str):
        _roles = [_roles]
    for _role in _roles:
        _role_id = _role if ":" in _role else None
        if not _role_id:
            if _node and _node_type:
                _role_id = get_role_id(_role, _node_type, _node)
            else:
                _role_id = get_role_id(_role)
        _roles_to_add.append(_role_id)

    # Construct the full payload
    _payload = {
        "data": {
            "type": "user",
            "roles_to_add": _roles_to_add
        }
    }

    # Perform and return the API call
    return api.put_request_with_retries(_uri, _payload, khoros_object=_khoros_object)


def assign_roles_to_user(khoros_object, user, lookup_type='id', roles_to_add=None, node=None, node_type='board',
                         v1=False, return_json=True):
    """This function assigns a user to one or more roles.

    .. versionadded:: 4.0.0

    :param khoros_object:_khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param user: The identifier (i.e. ID, login or email) of the user to be assigned to the role
    :type user: str
    :param lookup_type: The lookup type for the user identifier (``id``, ``login`` or ``email``)
    :type lookup_type: str
    :param roles_to_add: One or more roles (Role IDs or Role Names) to which the user will be assigned
    :type roles_to_add: str, list, tuple, set
    :param node: The Node ID of the node to which the role is scoped when applicable
    :type node: str, None
    :param node_type: The type of node to which the role is scoped (e.g. ``board`` (default), ``category``, etc.)
    :type node_type: str
    :param v1: Determines if the Community API v1 should be used to perform the operation (``False`` by default)
    :type v1: bool
    :param return_json: Determines if the response should be returned as JSON rather than XML (``True`` by default)
    :type return_json: bool
    :returns: The response from the API request
    :raises: :py:exc:`ValueError`, :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.CurrentlyUnsupportedError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`,
             :py:exc:`khoros.errors.exceptions.UnsupportedNodeTypeError`,
             :py:exc:`khoros.errors.exceptions.InvalidNodeTypeError`,
             :py:exc:`khoros.errors.exceptions.POSTRequestError`,
             :py:exc:`khoros.errors.exceptions.PUTRequestError`
    """
    # Validate the parameters
    if not roles_to_add:
        raise errors.exceptions.MissingRequiredDataError(param='roles_to_add')
    if lookup_type not in ['id', 'login', 'email']:
        raise errors.exceptions.InvalidLookupTypeError()
    node_type = _validate_node_type(node_type) if node else node_type

    # Call sub-functions depending on API version needed
    if v1:
        response = _assign_role_with_v1(khoros_object, user, lookup_type, roles_to_add, node, node_type, return_json)
    else:
        response = _assign_role_with_v2(khoros_object, user, lookup_type, roles_to_add, node, node_type)
    return response



# -*- coding: utf-8 -*-
"""
:Module:            khoros.structures.grouphubs
:Synopsis:          This module contains functions specific to group hubs within the Khoros Community platform
:Usage:             ``from khoros.structures import grouphubs``
:Example:           ``group_hub_url = grouphubs.create(khoros_object, gh_id, gh_title, disc_styles, return_url=True)``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     26 Dec 2020
"""

from .. import api, liql, errors
from ..utils import log_utils
from . import base

# Initialize the logger for this module
logger = log_utils.initialize_logging(__name__)

# Define the default discussion styles to use when creating new group hubs
DEFAULT_DISCUSSION_STYLES = ['blog', 'contest', 'forum', 'idea', 'qanda', 'tkb']

# Define a global variable to identify all discussion styles enabled for the environment where available
all_discussion_styles = DEFAULT_DISCUSSION_STYLES


def create(khoros_object, group_id, group_title, description=None, membership_type=None, open_group=None,
           closed_group=None, hidden_group=None, discussion_styles=None, enable_blog=None, enable_contest=None,
           enable_forum=None, enable_idea=None, enable_qanda=None, enable_tkb=None, all_styles_default=True,
           parent_category_id=None, avatar_image_path=None, full_response=None, return_id=None, return_url=None,
           return_api_url=None, return_http_code=None, return_status=None, return_error_messages=None,
           split_errors=False):
    """This function creates a new group hub within a Khoros Community environment.

    .. versionchanged:: 2.7.2
       Changed the data type for ``membership_type`` from ``dict`` to ``str`` in the docstring.

    .. versionadded:: 2.6.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param group_id: The unique identifier (i.e. ``id`` field) for the new group hub **(Required)**
    :type group_id: str, int
    :param group_title: The title of the group hub **(Required)**
    :type group_title: str
    :param description: A brief description of the group hub
    :type description: str, None
    :param membership_type: The ``membership_type`` value (``open``, ``closed`` or ``closed_hidden``)
    :type membership_type: str, None
    :param open_group: Defines the group hub as an open group
    :type open_group: bool, None
    :param closed_group: Defines the group hub as a closed group
    :type closed_group: bool, None
    :param hidden_group: Defines the group hub as a closed and hidden group
    :type hidden_group: bool, None
    :param discussion_styles: A list of discussion styles that will be permitted in the group hub
    :type discussion_styles: list, None
    :param enable_blog: Defines that the **blog** discussion style should be enabled for the group hub
    :type enable_blog: bool, None
    :param enable_contest: Defines that the **contest** discussion style should be enabled for the group hub
    :type enable_contest: bool, None
    :param enable_forum: Defines that the **forum** discussion style should be enabled for the group hub
    :type enable_forum: bool, None
    :param enable_idea: Defines that the **idea** discussion style should be enabled for the group hub
    :type enable_idea: bool, None
    :param enable_qanda: Defines that the **Q&A** (``qanda``) discussion style should be enabled for the group hub
    :type enable_qanda: bool, None
    :param enable_tkb: Defines that the **TKB** (``tkb``) discussion style should be enabled for the group hub
    :type enable_tkb: bool, None
    :param all_styles_default: Defines that all discussion styles should be enabled if not otherwise specified
    :type all_styles_default: bool
    :param parent_category_id: The parent category identifier (if applicable)
    :type parent_category_id: str, int, None
    :param avatar_image_path: The file path to the avatar image to be uploaded (if applicable)
    :type avatar_image_path: str, None
    :param full_response: Determines whether the full, raw API response should be returned by the function

                          .. caution:: This argument overwrites the ``return_id``, ``return_url``, ``return_api_url``,
                                       ``return_http_code``, ``return_status`` and ``return_error_messages`` arguments.

    :type full_response: bool, None
    :param return_id: Determines if the **ID** of the new group hub should be returned by the function
    :type return_id: bool, None
    :param return_url: Determines if the **URL** of the new group hub should be returned by the function
    :type return_url: bool, None
    :param return_api_url: Determines if the **API URL** of the new group hub should be returned by the function
    :type return_api_url: bool, None
    :param return_http_code: Determines if the **HTTP Code** of the API response should be returned by the function
    :type return_http_code: bool, None
    :param return_status: Determines if the **Status** of the API response should be returned by the function
    :type return_status: bool, None
    :param return_error_messages: Determines if any error messages associated with the API response should be
                                  returned by the function
    :type return_error_messages: bool, None
    :param split_errors: Defines whether or not error messages should be merged when applicable
    :type split_errors: bool
    :returns: Boolean value indicating a successful outcome (default), the full API response or one or more specific
              fields defined by function arguments
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`,
             :py:exc:`khoros.errors.exceptions.InvalidPayloadValueError`,
             :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.POSTRequestError`
    """
    payload = structure_payload(khoros_object, group_id, group_title, description, membership_type, open_group,
                                closed_group, hidden_group, discussion_styles, enable_blog, enable_contest,
                                enable_forum, enable_idea, enable_qanda, enable_tkb, all_styles_default,
                                parent_category_id)
    api_url = f"{khoros_object.core['v2_base']}/grouphubs"
    if avatar_image_path:
        response = _create_group_hub_with_avatar(khoros_object, api_url, payload, avatar_image_path)
    else:
        response = _create_group_hub_without_avatar(khoros_object, api_url, payload)
    return api.deliver_v2_results(response, full_response, return_id, return_url, return_api_url, return_http_code,
                                  return_status, return_error_messages, split_errors)


def _create_group_hub_with_avatar(_khoros_object, _api_url, _payload, _avatar_image_path):
    """This function creates a group hub with both a JSON payload and an image file to use as its avatar.

    :param _khoros_object: The core :py:class:`khoros.Khoros` object
    :type _khoros_object: class[khoros.Khoros]
    :param _api_url: The API URL to utilize in the API request
    :type _api_url: str
    :param _payload: The JSON payload to be used in the API request
    :type _payload: dict
    :param _avatar_image_path: The file path to the avatar image to be uploaded (if applicable)
    :type _avatar_image_path: str
    :returns: The API response from the multipart POST request
    :raises: :py:exc:`FileNotFoundError`,
             :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.POSTRequestError`
    """
    _headers = {'content-type': 'application/x-www-form-urlencoded'}
    _full_payload = api.combine_json_and_avatar_payload(_payload, _avatar_image_path)
    response = api.post_request_with_retries(_api_url, _full_payload, khoros_object=_khoros_object, multipart=True)
    return response


def _create_group_hub_without_avatar(_khoros_object, _api_url, _payload):
    """This function creates a group hub with only a JSON payload and no avatar image.

    .. versionadded:: 2.6.0

    :param _khoros_object: The core :py:class:`khoros.Khoros` object
    :type _khoros_object: class[khoros.Khoros]
    :param _api_url: The API URL to utilize in the API request
    :type _api_url: str
    :param _payload: The JSON payload to be used in the API request
    :type _payload: dict
    :returns: The API response from the POST request
    :raises: :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.POSTRequestError`
    """
    _headers = {'content-type': 'application/json'}
    _response = api.post_request_with_retries(_api_url, _payload, khoros_object=_khoros_object, headers=_headers)
    return _response


def structure_payload(khoros_object, group_id, group_title, description=None, membership_type=None, open_group=None,
                      closed_group=None, hidden_group=None, discussion_styles=None, enable_blog=None,
                      enable_contest=None, enable_forum=None, enable_idea=None, enable_qanda=None, enable_tkb=None,
                      all_styles_default=True, parent_category_id=None):
    """This function structures the payload to use in a Group Hub API request.

    .. versionchanged:: 2.7.3
       Changed the ``grouphub`` value in the initial ``payload`` definition to be a dictionary
       rather than a string to mitigate a :py:exc:`TypeError` exception getting raised.

    .. versionchanged:: 2.7.2
       Changed the data type for ``membership_type`` from ``dict`` to ``str`` in the docstring and
       fixed some bad logic raising false positive exceptions.

    .. versionadded:: 2.6.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param group_id: The unique identifier (i.e. ``id`` field) for the new group hub **(Required)**
    :type group_id: str, int
    :param group_title: The title of the group hub **(Required)**
    :type group_title: str
    :param description: A brief description of the group hub
    :type description: str, None
    :param membership_type: The ``membership_type`` value (``open``, ``closed`` or ``closed_hidden``)
    :type membership_type: str, None
    :param open_group: Defines the group hub as an open group
    :type open_group: bool, None
    :param closed_group: Defines the group hub as a closed group
    :type closed_group: bool, None
    :param hidden_group: Defines the group hub as a closed and hidden group
    :type hidden_group: bool, None
    :param discussion_styles: A list of discussion styles that will be permitted in the group hub
    :type discussion_styles: list, None
    :param enable_blog: Defines that the **blog** discussion style should be enabled for the group hub
    :type enable_blog: bool, None
    :param enable_contest: Defines that the **contest** discussion style should be enabled for the group hub
    :type enable_contest: bool, None
    :param enable_forum: Defines that the **forum** discussion style should be enabled for the group hub
    :type enable_forum: bool, None
    :param enable_idea: Defines that the **idea** discussion style should be enabled for the group hub
    :type enable_idea: bool, None
    :param enable_qanda: Defines that the **Q&A** (``qanda``) discussion style should be enabled for the group hub
    :type enable_qanda: bool, None
    :param enable_tkb: Defines that the **TKB** (``tkb``) discussion style should be enabled for the group hub
    :type enable_tkb: bool, None
    :param all_styles_default: Defines that all discussion styles should be enabled if not otherwise specified
    :type all_styles_default: bool
    :param parent_category_id: The parent category identifier (if applicable)
    :type parent_category_id: str, int, None
    :returns: The properly formatted payload for the API request
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`,
             :py:exc:`khoros.errors.exceptions.InvalidPayloadValueError`
    """
    payload = {"grouphub": {}}
    refresh_enabled_discussion_styles(khoros_object)
    required_error_msg = "The 'group_id', 'group_title' and 'membership_type' fields are required " \
                         "to create a group hub."
    if not all((khoros_object, group_id, group_title, membership_type)):
        raise errors.exceptions.MissingRequiredDataError(required_error_msg)
    payload = _structure_simple_string_fields(payload, group_id, group_title, description)
    payload = _structure_membership_type(payload, membership_type, open_group, closed_group, hidden_group)
    payload = _structure_discussion_styles(payload, discussion_styles, enable_blog, enable_contest, enable_forum,
                                           enable_idea, enable_qanda, enable_tkb, all_styles_default)
    payload = _structure_parent_category(payload, parent_category_id)
    return payload


def _structure_simple_string_fields(_payload, _group_id, _group_title, _description=None):
    """This function populates the payload with the group hub ID, title and description.

    .. versionadded:: 2.6.0

    :param _payload: The payload for a Group Hub API call
    :type _payload: dict
    :param _group_id: The unique identifier (ID) of the group hub
    :type _group_id: str
    :param _group_title: The title of the group hub
    :type _group_title: str
    :param _description: A brief description of the group hub
    :type _description: str, None
    :returns: The payload with the additional fields
    """
    _payload['grouphub']['id'] = str(_group_id)
    _payload['grouphub']['title'] = str(_group_title)
    if _description:
        _payload['grouphub']['description'] = _description
    return _payload


def _structure_membership_type(_payload, _membership_type, _open_group, _closed_group, _hidden_group):
    """This function populates the payload with the ``membership_type`` data.

    .. versionchanged:: 2.7.2
       Changed the data type for ``membership_type`` from ``dict`` to ``str`` in the docstring.

    .. versionadded:: 2.6.0

    :param _payload: The payload for a Group Hub API call
    :type _payload: dict
    :param _membership_type: The ``membership_type`` value (``open``, ``closed`` or ``closed_hidden``)
    :type _membership_type: str, None
    :param _open_group: Defines the group hub as an open group
    :type _open_group: bool, None
    :param _closed_group: Defines the group hub as a closed group
    :type _closed_group: bool, None
    :param _hidden_group: Defines the group hub as a closed and hidden group
    :type _hidden_group: bool, None
    :returns: The payload with the populated ``membership_type`` field
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    _valid_membership_types = ['open', 'closed', 'closed_hidden']
    _required_msg = "The membership type must be defined when creating a new group hub."
    if not any((_open_group, _closed_group, _hidden_group)) and _membership_type not in _valid_membership_types:
        raise errors.exceptions.MissingRequiredDataError(_required_msg)
    elif _membership_type and (_membership_type in _valid_membership_types):
        _payload['grouphub']['membership_type'] = _membership_type
    else:
        _types_and_values = {'open': _open_group, 'closed': _closed_group, 'closed_hidden': _hidden_group}
        for _type, _value in _types_and_values.items():
            if _value:
                _payload['grouphub']['membership_type'] = _type
                break
    return _payload


def _structure_discussion_styles(_payload, _discussion_styles=None, _enable_blog=None, _enable_contest=None,
                                 _enable_forum=None, _enable_idea=None, _enable_qanda=None, _enable_tkb=None,
                                 _all_styles_default=True):
    """This function defines the permitted discussion styles within the payload of a group hub API request.

    .. versionadded:: 2.6.0

    :param _payload: The payload to which the information should be added
    :type _payload: dict
    :param _discussion_styles: A list of discussion styles that will be permitted in the group hub
    :type _discussion_styles: list, None
    :param _enable_blog: Defines that the **blog** discussion style should be enabled for the group hub
    :type _enable_blog: bool, None
    :param _enable_contest: Defines that the **contest** discussion style should be enabled for the group hub
    :type _enable_contest: bool, None
    :param _enable_forum: Defines that the **forum** discussion style should be enabled for the group hub
    :type _enable_forum: bool, None
    :param _enable_idea: Defines that the **idea** discussion style should be enabled for the group hub
    :type _enable_idea: bool, None
    :param _enable_qanda: Defines that the **Q&A** (``qanda``) discussion style should be enabled for the group hub
    :type _enable_qanda: bool, None
    :param _enable_tkb: Defines that the **TKB** (``tkb``) discussion style should be enabled for the group hub
    :type _enable_tkb: bool, None
    :param _all_styles_default: Defines that all discussion styles should be enabled if not otherwise specified
    :type _all_styles_default: bool
    :returns: The payload with the included ``conversation_styles`` field and value(s)
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`,
             :py:exc:`khoros.errors.exceptions.InvalidPayloadValueError`
    """
    _required_msg = "At least one discussion style must be defined when creating a new group hub."
    if not any((_discussion_styles, _enable_blog, _enable_contest, _enable_forum, _enable_idea, _enable_qanda,
                _enable_tkb)):
        if _all_styles_default:
            _discussion_styles = all_discussion_styles
        else:
            raise errors.exceptions.MissingRequiredDataError(_required_msg)
    if _discussion_styles:
        if isinstance(_discussion_styles, str):
            _discussion_styles = [_discussion_styles]
        elif not isinstance(_discussion_styles, list):
            raise errors.exceptions.InvalidPayloadValueError(value=_discussion_styles, field='conversation_styles')
        for _style in _discussion_styles:
            if _style not in all_discussion_styles:
                raise errors.exceptions.InvalidPayloadValueError(value=_style, field='conversation_styles')
        _payload['grouphub']['conversation_styles'] = _discussion_styles
    else:
        _discussion_styles = []
        _discussion_toggles = {
            'blog': _enable_blog,
            'contest': _enable_contest,
            'forum': _enable_forum,
            'idea': _enable_idea,
            'qanda': _enable_qanda,
            'tkb': _enable_tkb
        }
        for _value, _toggle in _discussion_toggles.items():
            if _toggle:
                _discussion_styles.append(_value)
        _payload['grouphub']['conversation_styles'] = _discussion_styles
    _payload = _remove_disabled_discussion_styles(_payload)
    return _payload


def _remove_disabled_discussion_styles(_payload):
    """This function checks for any discussion styles that are disabled in the environment and removes them.

    :param _payload: The JSON payload to be used in an API request
    :type _payload: dict
    :returns: The payload with only enabled discussion styles
    """
    for _style in _payload['grouphub']['conversation_styles']:
        if _style not in all_discussion_styles:
            errors.handlers.eprint(f"The discussion style '{_style}' will be removed from the payload as it is a "
                                   "disabled discussion style in the environment.")
            _payload['grouphub']['conversation_styles'].remove(_style)
    return _payload


def _structure_parent_category(_payload, _parent_id):
    """This function structures the parent category field for a group hub if applicable.

    .. versionadded:: 2.6.0

    :param _payload: The payload to which the parent category field should be added
    :type _payload: dict
    :param _parent_id: The parent category identifier
    :type _parent_id: int, str, None
    :returns: The payload with the added field if applicable
    """
    if _parent_id:
        _parent_dict = {'id': str(_parent_id)}
        _payload['grouphub']['parent_category'] = _parent_dict
    return _payload


def get_total_count(khoros_object):
    """This function returns the total number of group hubs within the Khoros Community environment.

    .. versionadded:: 2.6.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :returns: The total number of group hubs as an integer
    """
    return liql.get_total_count(khoros_object, 'grouphubs')


def get_grouphub_id(url):
    """This function retrieves the Group Hub ID for a given group hub when provided its URL.

    .. versionadded:: 2.6.0

    :param url: The URL from which to parse out the Group Hub ID
    :type url: str
    :returns: The Group Hub ID retrieved from the URL
    :raises: :py:exc:`khoros.errors.exceptions.InvalidURLError`
    """
    return base.get_structure_id(url)


def grouphub_exists(khoros_object, grouphub_id=None, grouphub_url=None):
    """This function checks to see if a group hub exists.

    .. versionadded:: 2.7.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param grouphub_id: The ID of the group hub to check
    :type grouphub_id: str, None
    :param grouphub_url: The URL of the group hub to check
    :type grouphub_url: str, None
    :returns: Boolean value indicating whether or not the group hub already exists
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    return base.structure_exists(khoros_object, 'grouphub', grouphub_id, grouphub_url)


def refresh_enabled_discussion_styles(khoros_object):
    """This function refreshes the ``all_discussion_styles`` global variable to match what is in the
       core object settings when applicable.

    .. versionchanged:: 3.3.0
       Updated ``khoros_object._settings`` to be ``khoros_object.core_settings``.

    .. versionadded:: 2.6.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :returns: None
    """
    if 'discussion_styles' in khoros_object.core_settings:
        global all_discussion_styles
        all_discussion_styles = khoros_object.core_settings.get('discussion_styles')
    return


def _verify_group_hub_id(_group_hub_id, _group_hub_url):
    """This function verifies the Group Hub ID and looks it up as necessary using the Group Hub URL.

    .. versionadded:: 2.6.0

    :param _group_hub_id: The Group Hub ID supplied in the parent function
    :type _group_hub_id: str, None
    :param _group_hub_url: The Group Hub URL supplied in the parent function
    :type _group_hub_url: str, None
    :returns: The Group Hub ID
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    if not any((_group_hub_id, _group_hub_url)):
        raise errors.exceptions.MissingRequiredDataError("An ID or URL for the group hub must be provided.")
    if not _group_hub_id:
        _group_hub_id = get_grouphub_id(_group_hub_url)
    return _group_hub_id


def _structure_empty_payload():
    """This function returns the empty payload for a grouphub that is ready to be populated.

    .. versionadded:: 2.6.0

    :returns: The empty group hub payload as a dictionary
    """
    return {"grouphub": {}}


def update_title(khoros_object, new_title, group_hub_id=None, group_hub_url=None, full_response=None, return_id=None,
                 return_url=None, return_api_url=None, return_http_code=None, return_status=None,
                 return_error_messages=None, split_errors=False):
    """This function updates the title of an existing group hub.

    .. versionadded:: 2.6.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param new_title: The new title for the group hub
    :type new_title: str
    :param group_hub_id: The group hub ID that identifies the group hub to update (necessary if URL not provided)
    :type group_hub_id: str, None
    :param group_hub_url: The group hub URL that identifies the group hub to update (necessary if ID not provided)
    :type group_hub_url: str, None
    :param full_response: Determines whether the full, raw API response should be returned by the function

                          .. caution:: This argument overwrites the ``return_id``, ``return_url``, ``return_api_url``,
                                       ``return_http_code``, ``return_status`` and ``return_error_messages`` arguments.

    :type full_response: bool, None
    :param return_id: Determines if the **ID** of the new group hub should be returned by the function
    :type return_id: bool, None
    :param return_url: Determines if the **URL** of the new group hub should be returned by the function
    :type return_url: bool, None
    :param return_api_url: Determines if the **API URL** of the new group hub should be returned by the function
    :type return_api_url: bool, None
    :param return_http_code: Determines if the **HTTP Code** of the API response should be returned by the function
    :type return_http_code: bool, None
    :param return_status: Determines if the **Status** of the API response should be returned by the function
    :type return_status: bool, None
    :param return_error_messages: Determines if any error messages associated with the API response should be
                                  returned by the function
    :type return_error_messages: bool, None
    :param split_errors: Defines whether or not error messages should be merged when applicable
    :type split_errors: bool
    :returns: Boolean value indicating a successful outcome (default), the full API response or one or more specific
              fields defined by function arguments
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`,
             :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.PUTRequestError`
    """
    group_hub_id = _verify_group_hub_id(group_hub_id, group_hub_url)
    payload = _structure_empty_payload()
    payload['grouphub'] = {"title": new_title}
    api_url = f"{khoros_object.core['v2_base']}/grouphubs/{group_hub_id}"
    response = api.put_request_with_retries(api_url, payload, khoros_object=khoros_object)
    return api.deliver_v2_results(response, full_response, return_id, return_url, return_api_url, return_http_code,
                                  return_status, return_error_messages, split_errors)



# -*- coding: utf-8 -*-
"""
:Module:            khoros.errors.exceptions
:Synopsis:          Collection of exception classes relating to the khoros library
:Usage:             ``import khoros.errors.exceptions``
:Example:           ``raise khoros.errors.exceptions.BadCredentialsError()``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     07 Apr 2021
"""

#################
# Base Exception
#################


# Define base exception class
class KhorosError(Exception):
    """This is the base class for Khoros exceptions."""
    pass


############################
# Base Structure Exceptions
############################


class InvalidStructureTypeError(KhorosError):
    """This exception is used when an invalid node type is provided.

    .. versionadded:: 2.1.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The structure type that was provided is invalid."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'val' in kwargs:
            custom_msg = f"{default_msg.split('structure type ')[0]}'{kwargs['val']}'{default_msg.split('The')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


############################
# Authentication Exceptions
############################


class InvalidCallbackURLError(KhorosError):
    """This exception is used when an invalid Callback URL for OAuth 2.0 was not provided."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The OAuth 2.0 callback URL that was provided is invalid. The entire URL must be provided."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'val' in kwargs:
            custom_msg = f"{default_msg.split('URL ')[0]}'{kwargs['val']}'{default_msg.split('URL')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


class MissingAuthDataError(KhorosError):
    """This exception is used when authentication data is not supplied and therefore a connection cannot occur."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The authentication data was not provided and a connection cannot be established."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class SessionAuthenticationError(KhorosError):
    """This exception is used when the session key authentication attempt failed."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The session key authentication attempt failed."
        custom_msg = default_msg.replace('.', ' with the following message:')
        if not (args or kwargs):
            args = (default_msg,)
        elif 'message' in kwargs:
            custom_msg = f"{custom_msg} {kwargs['message']}"
            args = (custom_msg,)
        super().__init__(*args)


#####################
# General Exceptions
#####################


class CurrentlyUnsupportedError(KhorosError):
    """This exception is used when a feature or functionality being used is currently unsupported.

    .. versionchanged:: 2.0.0
       The unsupported feature can be passed as a string argument to explicitly reference it in the exception.
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "This feature is currently unsupported at this time."
        if not (args or kwargs):
            args = (default_msg,)
        else:
            custom_msg = f"The '{args[0]}' {default_msg.split('This ')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


class DataMismatchError(KhorosError):
    """This exception is used when there is a mismatch between two data sources.

    .. versionadded:: 2.3.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "A data mismatch was found with the data sources."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'data' in kwargs:
            multi_types = [list, tuple, set]
            if type(kwargs['data']) == str:
                custom_msg = f"{default_msg.split('data')[0]}'{kwargs['val']}'{default_msg.split('with the')[1]}"
                custom_msg = custom_msg.replace('sources', 'source')
                args = (custom_msg,)
            elif type(kwargs['data']) in multi_types and len(kwargs['data']) == 2:
                custom_section = f"'{kwargs['data'][0]}' and '{kwargs['data'][1]}'"
                custom_msg = f"{default_msg.split('data sources')[0]}{custom_section}{default_msg.split('with the')[1]}"
                args = (custom_msg,)
        super().__init__(*args)


class InvalidFieldError(KhorosError):
    """This exception is used when an invalid field is provided.

    .. versionadded:: 2.1.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The field that was provided is invalid."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'val' in kwargs:
            custom_msg = f"{default_msg.split('field ')[0]}'{kwargs['val']}'{default_msg.split('The')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


class InvalidURLError(KhorosError):
    """This exception is used when a provided URL is invalid.

    .. versionadded:: 2.1.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The provided URL is invalid"
        if not (args or kwargs):
            args = (default_msg,)
        elif 'url' in kwargs:
            custom_msg = f"{default_msg.split('is')[0]}'{kwargs['url']}'{default_msg.split('URL')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


class MissingRequiredDataError(KhorosError):
    """This exception is used when a function or method is missing one or more required arguments.

    .. versionchanged:: 4.0.0
       The exception can now accept the ``param`` keyword argument.

    .. versionadded:: 2.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "Missing one or more required parameters"
        init_msg = "The object failed to initialize as it is missing one or more required arguments."
        param_msg = "The required parameter 'PARAMETER_NAME' is not defined"
        if not (args or kwargs):
            args = (default_msg,)
        elif 'init' in args or 'initialize' in args:
            if 'object' in kwargs:
                custom_msg = f"{init_msg.split('object')[0]}'{kwargs['object']}'{init_msg.split('The')[1]}"
                args = (custom_msg,)
            else:
                args = (init_msg,)
        elif 'param' in kwargs:
            args = (param_msg.replace('PARAMETER_NAME', kwargs['param']),)
        else:
            args = (default_msg,)
        super().__init__(*args)


class UnknownFileTypeError(KhorosError):
    """This exception is used when a file type of a given file cannot be identified.

    .. versionadded:: 2.2.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The file type of the given file path cannot be identified."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'file' in kwargs:
            custom_msg = f"{default_msg.split('path')[0]}'{kwargs['file']}'{default_msg.split('path')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


#########################
# Generic API Exceptions
#########################


class APIConnectionError(KhorosError):
    """This exception is used when the API query could not be completed due to connection aborts and/or timeouts."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The API query could not be completed due to connection aborts and/or timeouts."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class APIRequestError(KhorosError):
    """This exception is used for generic API request errors when there isn't a more specific exception."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The DELETE request did not return a successful response."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class DELETERequestError(KhorosError):
    """This exception is used for generic DELETE request errors when there isn't a more specific exception."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The DELETE request did not return a successful response."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class FeatureNotConfiguredError(KhorosError):
    """This exception is used when an API request fails because a feature is not configured.

    .. versionadded:: 4.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        exc_msg = "The feature is not configured."
        if 'identifier' in kwargs or 'feature' in kwargs:
            if 'identifier' in kwargs:
                exc_msg += f" Identifier: {kwargs['identifier']}"
            if 'feature' in kwargs:
                exc_msg = exc_msg.replace("feature", f"{kwargs['feature']} feature")
            args = (exc_msg,)
        elif not (args or kwargs):
            args = (exc_msg,)
        super().__init__(*args)


class GETRequestError(KhorosError):
    """This exception is used for generic GET request errors when there isn't a more specific exception.

    .. versionchanged:: 3.2.0
       Enabled the ability to optionally pass ``status_code`` and/or ``message`` arguments.
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The GET request did not return a successful response."
        custom_msg = "The GET request failed with the following message:"
        if 'status_code' in kwargs or 'message' in kwargs:
            if 'status_code' in kwargs:
                status_code_msg = f"returned the {kwargs['status_code']} status code"
                custom_msg = custom_msg.replace('failed', status_code_msg)
            if 'message' in kwargs:
                custom_msg = f"{custom_msg} {kwargs['message']}"
            else:
                custom_msg = custom_msg.split(' with the following')[0] + "."
            args = (custom_msg,)
        elif not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class InvalidEndpointError(KhorosError):
    """This exception is used when an invalid API endpoint / service is provided."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The supplied endpoint for the API is not recognized. (Examples of valid " + \
                      "lookup types include 'people' and 'contents')"
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class InvalidLookupTypeError(KhorosError):
    """This exception is used when an invalid API lookup type is provided."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The supplied lookup type for the API is not recognized. (Examples of valid " + \
                      "lookup types include 'id' and 'email')"
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class InvalidPayloadValueError(KhorosError):
    """This exception is used when an invalid value is provided for a payload field.

    .. versionadded:: 2.6.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "An invalid payload value was provided."
        custom_msg = "The invalid payload value 'X' was provided for the 'Y' field."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'value' in kwargs:
            if 'field' in kwargs:
                custom_msg = custom_msg.replace('X', kwargs['value']).replace('Y', kwargs['field'])
            else:
                custom_msg = f"{custom_msg.replace('X', kwargs['value']).split(' for the')[0]}."
            args = (custom_msg,)
        super().__init__(*args)


class InvalidRequestTypeError(KhorosError):
    """This exception is used when an invalid API request type is provided."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The supplied request type for the API is not recognized. (Examples of valid " + \
                      "request types include 'POST' and 'PUT')"
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class LookupMismatchError(KhorosError):
    """This exception is used when an a lookup value doesn't match the supplied lookup type."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The supplied lookup type for the API does not match the value that was provided."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class NotFoundResponseError(KhorosError):
    """This exception is used when an API query returns a 404 response and there isn't a more specific class."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The API query returned a 404 response."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class PayloadMismatchError(KhorosError):
    """This exception is used when more than one payload is supplied for an API request.

    .. versionadded:: 3.2.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "More than one payload was provided for the API call when only one is permitted."
        if not (args or kwargs):
            args = (default_msg,)
        elif kwargs['request_type']:
            custom_msg = default_msg.replace("API call", f"{kwargs['request_type'].upper()} request")
            args = (custom_msg,)
        super().__init__(*args)


class POSTRequestError(KhorosError):
    """This exception is used for generic POST request errors when there isn't a more specific exception.

    .. versionchanged:: 3.2.0
       Enabled the ability to optionally pass ``status_code`` and/or ``message`` arguments.
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The POST request did not return a successful response."
        custom_msg = "The POST request failed with the following message:"
        if 'status_code' in kwargs or 'message' in kwargs:
            if 'status_code' in kwargs:
                status_code_msg = f"returned the {kwargs['status_code']} status code"
                custom_msg = custom_msg.replace('failed', status_code_msg)
            if 'message' in kwargs:
                custom_msg = f"{custom_msg} {kwargs['message']}"
            else:
                custom_msg = custom_msg.split(' with the following')[0] + "."
            args = (custom_msg,)
        elif not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class PUTRequestError(KhorosError):
    """This exception is used for generic PUT request errors when there isn't a more specific exception.

    .. versionchanged:: 3.2.0
       Enabled the ability to optionally pass ``status_code`` and/or ``message`` arguments.
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The PUT request did not return a successful response."
        custom_msg = "The PUT request failed with the following message:"
        if 'status_code' in kwargs or 'message' in kwargs:
            if 'status_code' in kwargs:
                status_code_msg = f"returned the {kwargs['status_code']} status code"
                custom_msg = custom_msg.replace('failed', status_code_msg)
            if 'message' in kwargs:
                custom_msg = f"{custom_msg} {kwargs['message']}"
            else:
                custom_msg = custom_msg.split(' with the following')[0] + "."
            args = (custom_msg,)
        elif not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


####################
# Helper Exceptions
####################


class InvalidHelperFileTypeError(KhorosError, ValueError):
    """This exception is used when an invalid file type is provided for the helper file."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The helper configuration file can only have the 'yaml' or 'json' file type."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class InvalidHelperArgumentsError(KhorosError):
    """This exception is used when the helper function was supplied arguments instead of keyword arguments."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The helper configuration file only accepts basic keyword arguments. (e.g. arg_name='arg_value')"
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class HelperFunctionNotFoundError(KhorosError):
    """This exception is used when a function referenced in the helper config file does not exist."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The function referenced in the helper configuration file could not be found."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


##################
# LiQL Exceptions
##################


class InvalidOperatorError(KhorosError):
    """This exception is used when an invalid operator is provided for the LiQL query."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "An invalid operator was provided for the LiQL query."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class LiQLParseError(KhorosError):
    """This exception is used when a function is unable to successfully parse a LiQL response.

    .. versionadded:: 3.2.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "Failed to parse the LiQL query response."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'message' in kwargs and kwargs['message']:
            custom_section = f" as the query failed with the following message: {kwargs['message']}"
            custom_msg = default_msg.replace('.', custom_section)
            args = (custom_msg,)
        super().__init__(*args)


class OperatorMismatchError(KhorosError):
    """This exception is used when the number of operators in the LiQL query does not match the number of fields."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The number of operators provided in the LiQL query does not match the number of fields/values."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class TooManyResultsError(KhorosError):
    """This exception is used when more results are returned than were expected in a LiQL query.

    .. versionchanged:: 3.2.0
       Fixed the default message to be appropriate as it was the same message found in another exception.

    .. versionadded:: 2.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "More results were returned in the LiQL response than were expected."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


#####################
# Message Exceptions
#####################


class MessageTypeNotFoundError(KhorosError):
    """This exception is used when a message type cannot be identified from a given URL."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The message type could not be identified in the provided URL."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'msg_type' in kwargs:
            custom_msg = f"{default_msg.split('message type ')[0]}'{kwargs['msg_type']}'{default_msg.split('type')[1]}"
            args = (custom_msg,)
        elif 'url' in kwargs:
            custom_msg = f"{default_msg.split('provided')[0]}following URL: {kwargs['url']}"
            args = (custom_msg,)
        super().__init__(*args)


##################
# Node Exceptions
##################


class InvalidNodeTypeError(KhorosError):
    """This exception is used when an invalid node type is provided."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The node type that was provided is invalid."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'val' in kwargs:
            custom_msg = f"{default_msg.split('node type ')[0]}'{kwargs['val']}'{default_msg.split('The')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


class NodeIDNotFoundError(KhorosError):
    """This exception is used when a valid Node ID could not be found in a provided URL."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "A valid Node ID could not be identified in the given URL."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'val' in kwargs:
            custom_msg = f"{default_msg.split('URL')[0]}: {kwargs['val']}"
            args = (custom_msg,)
        super().__init__(*args)


class NodeTypeNotFoundError(KhorosError):
    """This exception is used when a valid node type could not be found in a provided URL."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "A valid node type could not be identified in the given URL."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'val' in kwargs:
            custom_msg = f"{default_msg.split('URL')[0]}: {kwargs['val']}"
            args = (custom_msg,)
        super().__init__(*args)


class UnsupportedNodeTypeError(KhorosError):
    """This exception is used when an unsupported node type has been provided.

    .. versionadded:: 3.2.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The node type is unsupported with the given operation."
        if 'node_type' in kwargs:
            custom_msg = f"{default_msg.split('node ')[0]}'{kwargs['node_type']}' node{default_msg.split('node')[1]}"
            if 'operation' in kwargs:
                custom_msg = custom_msg.replace('the given operation', kwargs['operation'])
            args = (custom_msg,)
        elif not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


##################
# Role Exceptions
##################


class InvalidRoleError(KhorosError):
    """This exception is used when an invalid role is provided."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The role is invalid"
        if not (args or kwargs):
            args = (default_msg,)
        elif 'role' in kwargs:
            custom_msg = f"{default_msg.split('role ')[0]}'{kwargs['role']}'{default_msg.split('role')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


class InvalidRoleTypeError(KhorosError):
    """This exception is used when an invalid role type is provided."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The role type is invalid"
        if not (args or kwargs):
            args = (default_msg,)
        elif 'role_type' in kwargs:
            custom_msg = f"{default_msg.split('type ')[0]}'{kwargs['role_type']}'{default_msg.split('type ')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


##################
# User Exceptions
##################


class UserCreationError(KhorosError):
    """This exception is used when an attempt to create a user fails."""
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The user failed to be created."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'user' in kwargs:
            custom_msg = f"{default_msg.split('user ')[0]}'{kwargs['user']}'{default_msg.split('user')[1]}"
            args = (custom_msg,)
        if 'exc_msg' in kwargs:
            full_msg = f"{args[0].split('.')[0]} due to the following exception: {kwargs['exc_msg']}"
            args = (full_msg,)
        super().__init__(*args)



# -*- coding: utf-8 -*-
"""
:Module:         khoros.utils.tests
:Synopsis:       This package includes tests for the khoros library.
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  26 Mar 2020
"""


# -*- coding: utf-8 -*-
"""
:Module:            khoros.utils.tests.resources
:Synopsis:          Frequently used resources for performing unit testing
:Usage:             ``from khoros.utils.tests import resources``
:Example:           ``exceptions = resources.import_exceptions_module()``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     18 Jun 2020
:Version:           1.0.0
"""

import os
import sys
import importlib


def set_package_path():
    """This function adds the high-level khoros directory to the sys.path list.

    .. versionadded:: 2.7.4

    :returns: None
    """
    sys.path.insert(0, os.path.abspath('../..'))
    return


def import_modules(*modules):
    """This function imports and returns one or more modules to utilize in a unit test.

    .. versionadded:: 2.7.4

    :param modules: One or more module paths (absolute) in string format
    :returns: The imported module(s) as an individual object or a tuple of objects
    """
    imported_modules = []
    for module in modules:
        imported_modules.append(importlib.import_module(module))
    tuple(imported_modules)
    return imported_modules if len(imported_modules) > 1 else imported_modules[0]


def initialize_khoros_object():
    """This function imports the :py:class:`khoros.core.Khoros` class and initializes an object.

    .. versionadded:: 2.7.4

    :returns: The initialized :py:class:`khoros.core.Khoros` object
    """
    set_package_path()
    core_module = importlib.import_module('khoros.core')
    return core_module.Khoros(community_url='https://example.community.com', auto_connect=False,
                              tenant_id='example', auth_type='session_auth',
                              session_auth={'username': 'testuser', 'password': 'fakePassword123'})



# -*- coding: utf-8 -*-
"""
:Module:         khoros.utils.tests.test_liql_where_parsing
:Synopsis:       This module is used by pytest to verify that LiQL WHERE clauses can be parsed successfully.
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  11 Mar 2021
"""

import os
import sys


def set_package_path():
    """This function adds the high-level khoros directory to the sys.path list."""
    sys.path.insert(0, os.path.abspath('../..'))
    return


def parse_where_clauses():
    """This function runs through several parsing examples to ensure they all complete successfully.

    .. versionchanged:: 3.4.0
       Updated the name of the :py:func:`khoros.liql.parse_where_clause` function in its respective call.
    """
    # Import the liql module
    from khoros import liql

    # Define each test with a name
    examples = {
        '2-Length Tuple': ('id', 5),
        '3-Length Tuple': ('id', '>', 5),
        'Tuple of Tuples': (('id', 5), ('id', 6)),
        'List of Tuples': [('replies.count(*)', '>', 5), ('replies.count(*)', '<', 10)],
        'Logic Test': (('id', 5), ('id', 6), ('username', '!=', 'John')),
        'Set': {('id', 5), ('replies.count(*)', '=', 0)},
        '1-to-1 Dict': {'id': 5, 'replies.count(*)': 0},
        '1-to-2 Dict': {'id': ('>', 5), 'username': ('<', 10)}
    }

    # Perform each test
    for test_case, example in examples.items():
        # Define which logic to use per test
        if test_case == 'Logic Test':
            logic = ('AND', 'OR')
        else:
            logic = 'AND'

        # Convert any sets to lists for printing below without changing the original identifier
        if type(example) == set:
            print_example = list(example)
        else:
            print_example = example

        # Print results of each test
        print(f"Test Case:\t{test_case}")
        try:
            print(f"Length:\t\t{len(print_example)}")
            print(f"Length[0]:\t{len(print_example[0])}")
        except (KeyError, TypeError):
            pass
        print(f"Example:\t{print_example}")
        where = liql.parse_where_clause(example, logic)
        print(f"Parsed:\t\t{where}\n")
    return True


def test_where_clause_parsing():
    """This function tests to confirm that LiQL WHERE clauses are getting parsed properly without failing."""
    set_package_path()
    assert parse_where_clauses() is True        # nosec



# -*- coding: utf-8 -*-
"""
:Synopsis:          This script is the primary configuration file for the khoros project
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     07 Apr 2021
"""

import setuptools
import codecs
import os.path


def read(rel_path):
    """This function reads the ``version.py`` script in order to retrieve the version.

    .. versionadded:: 4.0.0
    """
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    """This function retrieves the current version of the package without needing to import the
       :py:mod:`khoros.utils.version` module in order to avoid dependency issues.

    .. versionadded:: 4.0.0
    """
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delimiter = '"' if '"' in line else "'"
            return line.split(delimiter)[1]
    raise RuntimeError("Unable to find the version string")


with open("README.md", "r") as fh:
    long_description = fh.read()

version = get_version("khoros/utils/version.py")

setuptools.setup(
    name="khoros",
    version=version,
    author="Jeff Shurtliff",
    author_email="jeff.shurtliff@rsa.com",
    description="Useful tools and utilities to assist in managing a Khoros Community (formerly Lithium) environment.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jeffshurtliff/khoros",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Communications",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Content Management System",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Message Boards",
        "Topic :: Internet :: WWW/HTTP :: Site Management",
        "Topic :: Office/Business",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    python_requires='>=3.6',
    install_requires=[
        "PyYAML>=5.3.1",
        "urllib3~=1.26.2",
        "requests>=2.23.0",
        "setuptools~=52.0.0",
    ],
)



# -*- coding: utf-8 -*-
"""
:Module:         khoros.utils.tests.test_core_utils
:Synopsis:       This module is used by pytest to verify that the core package utilities work properly
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  11 Mar 2021
"""

import os
import sys
import importlib


def set_package_path():
    """This function adds the high-level khoros directory to the sys.path list."""
    sys.path.insert(0, os.path.abspath('../..'))
    return


def import_core_utils():
    """This function imports the :py:mod:`khoros.utils.core_utils` module."""
    set_package_path()
    global core_utils
    core_utils = importlib.import_module('khoros.utils.core_utils')
    return


def test_url_encoding():
    """This function tests the :py:func:`khoros.utils.core_utils.url_encode` and
       :py:func:`khoros.utils.core_utils.url_decode` functions.
    """
    raw_url = "https://community.example.com/t5/bizapps/bizappspage/tab/community:studio:api-browser?searchString=select id from users where login = 'admin'"
    encoded_url = "https%3A%2F%2Fcommunity.example.com%2Ft5%2Fbizapps%2Fbizappspage%2Ftab%2Fcommunity%3Astudio%3Aapi-browser%3FsearchString%3Dselect+id+from+users+where+login+%3D+%27admin%27"
    assert core_utils.url_encode(raw_url) == encoded_url    # nosec
    assert core_utils.url_decode(encoded_url) == raw_url    # nosec
    return


def test_query_string_encoding():
    """This function tests the :py:func:`khoros.utils.core_utils.encode_query_string` function."""
    query_dict = {
        'user.login': 'exampleuser',
        'user.password': 'Ch@ngeM3!#$',
        'restapi.response_format': 'json',
    }
    encoded_string = 'user.login=exampleuser&user.password=Ch%40ngeM3%21%23%24&restapi.response_format=json'
    encoded_string_raw_pw = 'user.login=exampleuser&user.password=Ch@ngeM3!#$&restapi.response_format=json'
    assert core_utils.encode_query_string(query_dict) == encoded_string     # nosec
    assert core_utils.encode_query_string(query_dict, no_encode='user.password') == encoded_string_raw_pw   # nosec
    return


def test_numeric_eval():
    """This function tests the :py:func:`khoros.utils.core_utils.is_numeric` function."""
    numeric_string = '12345'
    integer = 12345
    non_numeric_string = 'ABcd1234'
    assert core_utils.is_numeric(numeric_string) and core_utils.is_numeric(integer)     # nosec
    assert not core_utils.is_numeric(non_numeric_string)    # nosec
    return


def test_remove_tld():
    """This function tests to the :py:func:`khoros.utils.core_utils.remove_tld` function."""
    url = 'https://community.khoros.com/t5/Developer-Discussion/Add-multiple-tags-to-a-message-with-a-' \
          'single-API-call/m-p/596532#M17061'
    control_data = {
        'with': '/t5/Developer-Discussion/Add-multiple-tags-to-a-message-with-a-single-API-call/m-p/596532#M17061',
        'without': '/t5/Developer-Discussion/Add-multiple-tags-to-a-message-with-a-single-API-call/m-p/596532'
    }
    assert core_utils.remove_tld(url, strip_anchors=False) == control_data.get('with')      # nosec
    assert core_utils.remove_tld(url, strip_anchors=True) == control_data.get('without')    # nosec
    return


def test_merge_and_dedup():
    """This function tests the :py:func:`khoros.utils.core_utils.merge_and_dedup` function."""
    control_data = ['one', 'two', 'three', 'four', 'five', 6]
    test_data = core_utils.merge_and_dedup('one', 'two', 'two', ('two', 'three'), ['four', 'four'],
                                           {'four', 'five'}, 6, {6})
    assert test_data == control_data    # nosec
    return


def _check_type_and_items(_converted_item, _control_item, _new_type):
    """This function facilitates testing the :py:func:`khoros.utils.core_utils.convert_set` function."""
    _correct_type = True if isinstance(_converted_item, _new_type) else False
    _items_present = True
    for _item in _control_item:
        if _item not in _converted_item:
            _items_present = False
    return all((_correct_type, _items_present))


def test_convert_set():
    """This function tests the :py:func:`khoros.utils.core_utils.convert_set` function."""
    some_set = {'test', 1, 'two', ('three', )}
    some_list = ['test', 1, 'two', ('three', )]
    some_tuple = ('test', 1, 'two', ('three',))
    assert _check_type_and_items(core_utils.convert_set(some_set), some_list, list)                 # nosec
    assert _check_type_and_items(core_utils.convert_set(some_set, 'tuple'), some_tuple, tuple)      # nosec
    assert _check_type_and_items(core_utils.convert_set(some_list), some_list, list)                # nosec
    assert _check_type_and_items(core_utils.convert_set(some_tuple), some_tuple, tuple)             # nosec
    assert _check_type_and_items(core_utils.convert_set('some string'), 'some string', str)         # nosec
    return


# Import the khoros.utils.core_utils module
import_core_utils()



# -*- coding: utf-8 -*-
"""
:Module:            khoros.objects
:Synopsis:          This module contains sub-modules that are specific to the various Community API objects.
:Usage:             ``from khoros import objects``
:Example:           ``objects.archives.archive(khoros_obj, '123', suggested_url, return_status=True)``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     25 Mar 2021
"""

__all__ = ['albums', 'archives', 'messages', 'roles', 'settings', 'subscriptions', 'users']

# Import all submodules by default
from . import albums
from . import archives
from . import messages
from . import roles
from . import settings
from . import subscriptions
from . import users
